@@ -1,15 +1,22 @@
 // Copyright (C) 2021 The Qt Company Ltd.
 // SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
-#include "qvideotexturehelper_p.h"
-#include "qabstractvideobuffer_p.h"
-#include "qvideoframeconverter_p.h"
+#include "qabstractvideobuffer.h"
 
+#include "qvideotexturehelper_p.h"
+#include "qvideoframeconverter_p.h"
+#include "qvideoframe_p.h"
+#include "qvideoframetexturefromsource_p.h"
+#include "private/qmultimediautils_p.h"
+
+#include <QtCore/qfile.h>
 #include <qpainter.h>
 #include <qloggingcategory.h>
 
 QT_BEGIN_NAMESPACE
 
+Q_STATIC_LOGGING_CATEGORY(qLcVideoTextureHelper, "qt.multimedia.video.texturehelper")
+
 namespace QVideoTextureHelper
 {
 
@@ -17,121 +24,121 @@ static const TextureDescription descriptions[QVideoFrameFormat::NPixelFormats] =
     //  Format_Invalid
     { 0, 0,
       [](int, int) { return 0; },
-     { QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat},
+     { TextureDescription::UnknownFormat, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat},
      { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_ARGB8888
     { 1, 4,
         [](int stride, int height) { return stride*height; },
-        { QRhiTexture::RGBA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+        { TextureDescription::RGBA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
         { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_ARGB8888_Premultiplied
     { 1, 4,
         [](int stride, int height) { return stride*height; },
-        { QRhiTexture::RGBA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+        { TextureDescription::RGBA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
         { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_XRGB8888
     { 1, 4,
         [](int stride, int height) { return stride*height; },
-        { QRhiTexture::RGBA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+        { TextureDescription::RGBA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
         { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_BGRA8888
     { 1, 4,
         [](int stride, int height) { return stride*height; },
-        { QRhiTexture::BGRA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+        { TextureDescription::BGRA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
         { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_BGRA8888_Premultiplied
     { 1, 4,
         [](int stride, int height) { return stride*height; },
-        { QRhiTexture::BGRA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+        { TextureDescription::BGRA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
         { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_BGRX8888
     { 1, 4,
         [](int stride, int height) { return stride*height; },
-        { QRhiTexture::BGRA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+        { TextureDescription::BGRA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
         { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_ABGR8888
     { 1, 4,
         [](int stride, int height) { return stride*height; },
-        { QRhiTexture::RGBA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+        { TextureDescription::RGBA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
         { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_XBGR8888
     { 1, 4,
         [](int stride, int height) { return stride*height; },
-        { QRhiTexture::RGBA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+        { TextureDescription::RGBA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
         { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_RGBA8888
     { 1, 4,
         [](int stride, int height) { return stride*height; },
-        { QRhiTexture::RGBA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+        { TextureDescription::RGBA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
         { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_RGBX8888
     { 1, 4,
         [](int stride, int height) { return stride*height; },
-        { QRhiTexture::RGBA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+        { TextureDescription::RGBA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
         { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_AYUV
     { 1, 4,
       [](int stride, int height) { return stride*height; },
-     { QRhiTexture::RGBA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+     { TextureDescription::RGBA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
      { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_AYUV_Premultiplied
     { 1, 4,
         [](int stride, int height) { return stride*height; },
-        { QRhiTexture::RGBA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+        { TextureDescription::RGBA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
         { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_YUV420P
     { 3, 1,
       [](int stride, int height) { return stride * ((height * 3 / 2 + 1) & ~1); },
-     { QRhiTexture::R8, QRhiTexture::R8, QRhiTexture::R8 },
+     { TextureDescription::Red_8, TextureDescription::Red_8, TextureDescription::Red_8 },
      { { 1, 1 }, { 2, 2 }, { 2, 2 } }
     },
      // Format_YUV422P
     { 3, 1,
       [](int stride, int height) { return stride * height * 2; },
-     { QRhiTexture::R8, QRhiTexture::R8, QRhiTexture::R8 },
+     {TextureDescription::Red_8, TextureDescription::Red_8, TextureDescription::Red_8 },
      { { 1, 1 }, { 2, 1 }, { 2, 1 } }
     },
      // Format_YV12
     { 3, 1,
       [](int stride, int height) { return stride * ((height * 3 / 2 + 1) & ~1); },
-     { QRhiTexture::R8, QRhiTexture::R8, QRhiTexture::R8 },
+     {TextureDescription::Red_8, TextureDescription::Red_8, TextureDescription::Red_8 },
      { { 1, 1 }, { 2, 2 }, { 2, 2 } }
     },
     // Format_UYVY
     { 1, 2,
       [](int stride, int height) { return stride*height; },
-     { QRhiTexture::RGBA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+     { TextureDescription::RGBA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
      { { 2, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_YUYV
     { 1, 2,
       [](int stride, int height) { return stride*height; },
-     { QRhiTexture::RGBA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+     { TextureDescription::RGBA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
      { { 2, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_NV12
     { 2, 1,
       [](int stride, int height) { return stride * ((height * 3 / 2 + 1) & ~1); },
-     { QRhiTexture::R8, QRhiTexture::RG8, QRhiTexture::UnknownFormat },
+     { TextureDescription::Red_8, TextureDescription::RG_8, TextureDescription::UnknownFormat },
      { { 1, 1 }, { 2, 2 }, { 1, 1 } }
     },
     // Format_NV21
     { 2, 1,
       [](int stride, int height) { return stride * ((height * 3 / 2 + 1) & ~1); },
-     { QRhiTexture::R8, QRhiTexture::RG8, QRhiTexture::UnknownFormat },
+     { TextureDescription::Red_8, TextureDescription::RG_8, TextureDescription::UnknownFormat },
      { { 1, 1 }, { 2, 2 }, { 1, 1 } }
     },
     // Format_IMC1
@@ -142,13 +149,13 @@ static const TextureDescription descriptions[QVideoFrameFormat::NPixelFormats] =
           h += 2*(((h/2) + 15) & ~15);
           return stride * h;
       },
-     { QRhiTexture::R8, QRhiTexture::R8, QRhiTexture::R8 },
+     {TextureDescription::Red_8,TextureDescription::Red_8,TextureDescription::Red_8 },
      { { 1, 1 }, { 2, 2 }, { 2, 2 } }
     },
     // Format_IMC2
     { 2, 1,
       [](int stride, int height) { return 2*stride*height; },
-     { QRhiTexture::R8, QRhiTexture::R8, QRhiTexture::UnknownFormat },
+     {TextureDescription::Red_8,TextureDescription::Red_8, TextureDescription::UnknownFormat },
      { { 1, 1 }, { 1, 2 }, { 1, 1 } }
     },
     // Format_IMC3
@@ -159,67 +166,130 @@ static const TextureDescription descriptions[QVideoFrameFormat::NPixelFormats] =
           h += 2*(((h/2) + 15) & ~15);
           return stride * h;
       },
-     { QRhiTexture::R8, QRhiTexture::R8, QRhiTexture::R8 },
+     {TextureDescription::Red_8,TextureDescription::Red_8,TextureDescription::Red_8 },
      { { 1, 1 }, { 2, 2 }, { 2, 2 } }
     },
     // Format_IMC4
     { 2, 1,
       [](int stride, int height) { return 2*stride*height; },
-     { QRhiTexture::R8, QRhiTexture::R8, QRhiTexture::UnknownFormat },
+     {TextureDescription::Red_8,TextureDescription::Red_8, TextureDescription::UnknownFormat },
      { { 1, 1 }, { 1, 2 }, { 1, 1 } }
     },
     // Format_Y8
     { 1, 1,
       [](int stride, int height) { return stride*height; },
-     { QRhiTexture::R8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+     {TextureDescription::Red_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
      { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_Y16
     { 1, 2,
       [](int stride, int height) { return stride*height; },
-     { QRhiTexture::R16, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+     { TextureDescription::Red_16, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
      { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_P010
     { 2, 2,
       [](int stride, int height) { return stride * ((height * 3 / 2 + 1) & ~1); },
-     { QRhiTexture::R16, QRhiTexture::RG16, QRhiTexture::UnknownFormat },
+     { TextureDescription::Red_16, TextureDescription::RG_16, TextureDescription::UnknownFormat },
      { { 1, 1 }, { 2, 2 }, { 1, 1 } }
     },
     // Format_P016
     { 2, 2,
       [](int stride, int height) { return stride * ((height * 3 / 2 + 1) & ~1); },
-     { QRhiTexture::R16, QRhiTexture::RG16, QRhiTexture::UnknownFormat },
+     { TextureDescription::Red_16, TextureDescription::RG_16, TextureDescription::UnknownFormat },
      { { 1, 1 }, { 2, 2 }, { 1, 1 } }
     },
     // Format_SamplerExternalOES
     {
         1, 0,
         [](int, int) { return 0; },
-        { QRhiTexture::RGBA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+        { TextureDescription::RGBA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
         { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_Jpeg
     { 1, 4,
       [](int stride, int height) { return stride*height; },
-     { QRhiTexture::RGBA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+     { TextureDescription::RGBA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
      { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_SamplerRect
     {
         1, 0,
         [](int, int) { return 0; },
-        { QRhiTexture::BGRA8, QRhiTexture::UnknownFormat, QRhiTexture::UnknownFormat },
+        { TextureDescription::BGRA_8, TextureDescription::UnknownFormat, TextureDescription::UnknownFormat },
         { { 1, 1 }, { 1, 1 }, { 1, 1 } }
     },
     // Format_YUV420P10
-    { 3, 1,
+    { 3, 2,
         [](int stride, int height) { return stride * ((height * 3 / 2 + 1) & ~1); },
-        { QRhiTexture::R16, QRhiTexture::R16, QRhiTexture::R16 },
+        { TextureDescription::Red_16, TextureDescription::Red_16, TextureDescription::Red_16 },
         { { 1, 1 }, { 2, 2 }, { 2, 2 } }
     },
 };
 
+Q_GLOBAL_STATIC(QList<QRhiTexture::Format>, g_excludedRhiTextureFormats) // for tests only
+
+static bool isRhiTextureFormatSupported(const QRhi *rhi, QRhiTexture::Format format)
+{
+    if (g_excludedRhiTextureFormats->contains(format))
+        return false;
+    if (!rhi) // consider the format is supported if no rhi specified
+        return true;
+    return rhi->isTextureFormatSupported(format);
+}
+
+static QRhiTexture::Format
+resolveRhiTextureFormat(QRhi *rhi, QRhiTexture::Format format,
+                        std::initializer_list<QRhiTexture::Format> fallbacks)
+{
+    if (isRhiTextureFormatSupported(rhi, format))
+        return format;
+
+    for (QRhiTexture::Format fallback : fallbacks) {
+        if (!isRhiTextureFormatSupported(rhi, fallback))
+            continue;
+
+        qCDebug(qLcVideoTextureHelper) << "Using fallback texture format" << fallback;
+        return fallback;
+    }
+
+    // TODO: QTBUG-135911: In some cases rhi claims format and fallbacks are all
+    // unsupported, but when using preferred format video plays fine
+    qCDebug(qLcVideoTextureHelper) << "Cannot determine any usable texture format, using preferred format" << format;
+    return format;
+}
+
+QRhiTexture::Format TextureDescription::rhiTextureFormat(int plane, QRhi *rhi) const
+{
+    switch (textureFormat[plane]) {
+        case UnknownFormat:
+            return QRhiTexture::UnknownFormat;
+        case Red_8:
+            // NOTE: RED_OR_ALPHA8 requires special alpha shaders if rhi doesn't have feature
+            // RedOrAlpha8IsRed
+            return resolveRhiTextureFormat(rhi, QRhiTexture::R8, { QRhiTexture::RED_OR_ALPHA8 });
+        case RG_8:
+            return resolveRhiTextureFormat(rhi, QRhiTexture::RG8, { QRhiTexture::RGBA8 });
+        case RGBA_8:
+            return resolveRhiTextureFormat(rhi, QRhiTexture::RGBA8, {});
+        case BGRA_8:
+            return resolveRhiTextureFormat(rhi, QRhiTexture::BGRA8, {});
+        case Red_16:
+            // TODO: Special handling for 16-bit formats, if we want to support them at all.
+            // Otherwise should give an error.
+            return resolveRhiTextureFormat(rhi, QRhiTexture::R16, { QRhiTexture::RG8 });
+        case RG_16:
+            return resolveRhiTextureFormat(rhi, QRhiTexture::RG16, { QRhiTexture::RGBA8 });
+        default:
+            Q_UNREACHABLE();
+        }
+}
+
+void setExcludedRhiTextureFormats(QList<QRhiTexture::Format> formats)
+{
+    g_excludedRhiTextureFormats->swap(formats);
+}
+
 const TextureDescription *textureDescription(QVideoFrameFormat::PixelFormat format)
 {
     return descriptions + format;
@@ -242,88 +312,91 @@ QString vertexShaderFileName(const QVideoFrameFormat &format)
     return QStringLiteral(":/qt-project.org/multimedia/shaders/vertex.vert.qsb");
 }
 
-QString fragmentShaderFileName(const QVideoFrameFormat &format, QRhiSwapChain::Format surfaceFormat)
+QString fragmentShaderFileName(const QVideoFrameFormat &format, QRhi *,
+                               QRhiSwapChain::Format surfaceFormat)
 {
-    const char *shader = nullptr;
+    QString shaderFile;
     switch (format.pixelFormat()) {
     case QVideoFrameFormat::Format_Y8:
+        shaderFile = QStringLiteral("y");
+        break;
     case QVideoFrameFormat::Format_Y16:
-        shader = "y";
+        shaderFile = QStringLiteral("y16");
         break;
     case QVideoFrameFormat::Format_AYUV:
     case QVideoFrameFormat::Format_AYUV_Premultiplied:
-        shader = "ayuv";
+        shaderFile = QStringLiteral("ayuv");
         break;
     case QVideoFrameFormat::Format_ARGB8888:
     case QVideoFrameFormat::Format_ARGB8888_Premultiplied:
     case QVideoFrameFormat::Format_XRGB8888:
-        shader = "argb";
+        shaderFile = QStringLiteral("argb");
         break;
     case QVideoFrameFormat::Format_ABGR8888:
     case QVideoFrameFormat::Format_XBGR8888:
-        shader = "abgr";
+        shaderFile = QStringLiteral("abgr");
         break;
     case QVideoFrameFormat::Format_Jpeg: // Jpeg is decoded transparently into an ARGB texture
-        shader = "bgra";
+        shaderFile = QStringLiteral("bgra");
         break;
     case QVideoFrameFormat::Format_RGBA8888:
     case QVideoFrameFormat::Format_RGBX8888:
     case QVideoFrameFormat::Format_BGRA8888:
     case QVideoFrameFormat::Format_BGRA8888_Premultiplied:
     case QVideoFrameFormat::Format_BGRX8888:
-        shader = "rgba";
+        shaderFile = QStringLiteral("rgba");
         break;
     case QVideoFrameFormat::Format_YUV420P:
     case QVideoFrameFormat::Format_YUV422P:
     case QVideoFrameFormat::Format_IMC3:
-        shader = "yuv_triplanar";
+        shaderFile = QStringLiteral("yuv_triplanar");
         break;
     case QVideoFrameFormat::Format_YUV420P10:
-        shader = "yuv_triplanar_p10";
+        shaderFile = QStringLiteral("yuv_triplanar_p10");
         break;
     case QVideoFrameFormat::Format_YV12:
     case QVideoFrameFormat::Format_IMC1:
-        shader = "yvu_triplanar";
+        shaderFile = QStringLiteral("yvu_triplanar");
         break;
     case QVideoFrameFormat::Format_IMC2:
-        shader = "imc2";
+        shaderFile = QStringLiteral("imc2");
         break;
     case QVideoFrameFormat::Format_IMC4:
-        shader = "imc4";
+        shaderFile = QStringLiteral("imc4");
         break;
     case QVideoFrameFormat::Format_UYVY:
-        shader = "uyvy";
+        shaderFile = QStringLiteral("uyvy");
         break;
     case QVideoFrameFormat::Format_YUYV:
-        shader = "yuyv";
+        shaderFile = QStringLiteral("yuyv");
         break;
     case QVideoFrameFormat::Format_P010:
     case QVideoFrameFormat::Format_P016:
         // P010/P016 have the same layout as NV12, just 16 instead of 8 bits per pixel
         if (format.colorTransfer() == QVideoFrameFormat::ColorTransfer_ST2084) {
-            shader = "nv12_bt2020_pq";
+            shaderFile = QStringLiteral("nv12_bt2020_pq");
             break;
         }
         if (format.colorTransfer() == QVideoFrameFormat::ColorTransfer_STD_B67) {
-            shader = "nv12_bt2020_hlg";
+            shaderFile = QStringLiteral("nv12_bt2020_hlg");
             break;
         }
-        // Fall through, should be bt709
-        Q_FALLTHROUGH();
+        shaderFile = QStringLiteral("p016");
+        break;
     case QVideoFrameFormat::Format_NV12:
-        shader = "nv12";
+        shaderFile = QStringLiteral("nv12");
         break;
     case QVideoFrameFormat::Format_NV21:
-        shader = "nv21";
+        shaderFile = QStringLiteral("nv21");
         break;
     case QVideoFrameFormat::Format_SamplerExternalOES:
 #if 1//def Q_OS_ANDROID
-        shader = "externalsampler";
+        shaderFile = QStringLiteral("externalsampler");
         break;
 #endif
     case QVideoFrameFormat::Format_SamplerRect:
 #if 1//def Q_OS_MACOS
-        shader = "rectsampler_bgra";
+        shaderFile = QStringLiteral("rectsampler_bgra");
         break;
 #endif
         // fallthrough
@@ -331,12 +404,19 @@ QString fragmentShaderFileName(const QVideoFrameFormat &format, QRhiSwapChain::F
     default:
         break;
     }
-    if (!shader)
+
+    if (shaderFile.isEmpty())
         return QString();
-    QString shaderFile = QStringLiteral(":/qt-project.org/multimedia/shaders/") + QString::fromLatin1(shader);
+
+    shaderFile.prepend(u":/qt-project.org/multimedia/shaders/");
+
     if (surfaceFormat == QRhiSwapChain::HDRExtendedSrgbLinear)
-        shaderFile += QLatin1String("_linear");
-    shaderFile += QStringLiteral(".frag.qsb");
+        shaderFile.append(u"_linear");
+
+    shaderFile.append(u".frag.qsb");
+
+    Q_ASSERT_X(QFile::exists(shaderFile), Q_FUNC_INFO,
+               QStringLiteral("Shader file %1 does not exist").arg(shaderFile).toLatin1());
     return shaderFile;
 }
 
@@ -474,7 +554,9 @@ static float convertSDRFromLinear(float sig)
     return sig;
 }
 
-void updateUniformData(QByteArray *dst, const QVideoFrameFormat &format, const QVideoFrame &frame, const QMatrix4x4 &transform, float opacity, float maxNits)
+void updateUniformData(QByteArray *dst, QRhi *rhi, const QVideoFrameFormat &format,
+                       const QVideoFrame &frame, const QMatrix4x4 &transform, float opacity,
+                       float maxNits)
 {
 #ifndef Q_OS_ANDROID
     Q_UNUSED(frame);
@@ -520,7 +602,8 @@ void updateUniformData(QByteArray *dst, const QVideoFrameFormat &format, const Q
         break;
     case QVideoFrameFormat::Format_SamplerExternalOES:
         // get Android specific transform for the externalsampler texture
-        cmat = frame.videoBuffer()->externalTextureMatrix();
+        if (auto hwBuffer = QVideoFramePrivate::hwBuffer(frame))
+            cmat = hwBuffer->externalTextureMatrix();
         break;
     case QVideoFrameFormat::Format_SamplerRect:
     {
@@ -559,112 +642,120 @@ void updateUniformData(QByteArray *dst, const QVideoFrameFormat &format, const Q
     ud->width = float(format.frameWidth());
     ud->masteringWhite = fromLinear(float(format.maxLuminance())/100.f);
     ud->maxLum = fromLinear(float(maxNits)/100.f);
+    const TextureDescription* desc = textureDescription(format.pixelFormat());
+
+    // Let's consider using the red component if Red_8 is not used,
+    // it's useful for compatibility the shaders with 16bit formats.
+
+    const bool useRedComponent =
+            !desc->hasTextureFormat(TextureDescription::Red_8) ||
+            isRhiTextureFormatSupported(rhi, QRhiTexture::R8) ||
+            rhi->isFeatureSupported(QRhi::RedOrAlpha8IsRed);
+    ud->redOrAlphaIndex = useRedComponent ? 0 : 3; // r:0 g:1 b:2 a:3
+    for (int plane = 0; plane < desc->nplanes; ++plane)
+        ud->planeFormats[plane] = desc->rhiTextureFormat(plane, rhi);
 }
 
-// There's no maping between QRhiTexture formats and QImage formats.
-// The function gets a size compatible format; it works fine as
-// QRhi relies on the given QRhiTexture format instead of the image format.
-static QImage::Format sizeCompatibleImageFormat(QRhiTexture::Format rhiFromat)
-{
-    switch (rhiFromat) {
-    case QRhiTexture::RGBA8:
-    case QRhiTexture::BGRA8:
-    case QRhiTexture::RG16:
-        return QImage::Format_ARGB32;
-    case QRhiTexture::R16:
-    case QRhiTexture::RG8:
-        return QImage::Format_Grayscale16;
-    case QRhiTexture::R8:
-        return QImage::Format_Grayscale8;
-    default:
-        qWarning() << "QRhiTexture::Format" << rhiFromat
-                   << "must be mapped to a size-compatible QImage::Format";
-        Q_ASSERT(!"QRhiTexture::Format is not mapped to a size-compatible QImage::Format");
-        return QImage::Format_Grayscale8;
-    }
-}
+enum class UpdateTextureWithMapResult : uint8_t {
+    Failed,
+    UpdatedWithDataCopy,
+    UpdatedWithDataReference
+};
 
-static bool updateTextureWithMap(QVideoFrame& frame, QRhi *rhi, QRhiResourceUpdateBatch *rub, int plane, std::unique_ptr<QRhiTexture> &tex)
+static UpdateTextureWithMapResult updateTextureWithMap(const QVideoFrame &frame, QRhi &rhi,
+                                                       QRhiResourceUpdateBatch &rub, int plane,
+                                                       std::unique_ptr<QRhiTexture> &tex)
 {
+    Q_ASSERT(frame.isMapped());
+
     QVideoFrameFormat fmt = frame.surfaceFormat();
     QVideoFrameFormat::PixelFormat pixelFormat = fmt.pixelFormat();
     QSize size = fmt.frameSize();
 
     const TextureDescription &texDesc = descriptions[pixelFormat];
-    QSize planeSize(size.width()/texDesc.sizeScale[plane].x, size.height()/texDesc.sizeScale[plane].y);
+    QSize planeSize = texDesc.rhiPlaneSize(size, plane, &rhi);
 
-    bool needsRebuild = !tex || tex->pixelSize() != planeSize || tex->format() != texDesc.textureFormat[plane];
+    bool needsRebuild = !tex || tex->pixelSize() != planeSize || tex->format() != texDesc.rhiTextureFormat(plane, &rhi);
     if (!tex) {
-        tex.reset(rhi->newTexture(texDesc.textureFormat[plane], planeSize, 1, {}));
+        tex.reset(rhi.newTexture(texDesc.rhiTextureFormat(plane, &rhi), planeSize, 1, {}));
         if (!tex) {
             qWarning("Failed to create new texture (size %dx%d)", planeSize.width(), planeSize.height());
-            return false;
+            return UpdateTextureWithMapResult::Failed;
         }
     }
 
     if (needsRebuild) {
-        tex->setFormat(texDesc.textureFormat[plane]);
+        tex->setFormat(texDesc.rhiTextureFormat(plane, &rhi));
         tex->setPixelSize(planeSize);
         if (!tex->create()) {
             qWarning("Failed to create texture (size %dx%d)", planeSize.width(), planeSize.height());
-            return false;
+            return UpdateTextureWithMapResult::Failed;
         }
     }
 
+    auto result = UpdateTextureWithMapResult::UpdatedWithDataCopy;
+
     QRhiTextureSubresourceUploadDescription subresDesc;
-    QImage image;
+
     if (pixelFormat == QVideoFrameFormat::Format_Jpeg) {
         Q_ASSERT(plane == 0);
 
+        QImage image;
+
         // calling QVideoFrame::toImage is not accurate. To be fixed.
-        image = frame.toImage();
+        // frame transformation will be considered later
+        const QVideoFrameFormat surfaceFormat = frame.surfaceFormat();
+
+        const bool hasSurfaceTransform = surfaceFormat.isMirrored()
+                || surfaceFormat.scanLineDirection() == QVideoFrameFormat::BottomToTop
+                || surfaceFormat.rotation() != QtVideo::Rotation::None;
+
+        if (hasSurfaceTransform)
+            image = qImageFromVideoFrame(frame, VideoTransformation{});
+        else
+            image = frame.toImage(); // use the frame cache, no surface transforms applied
+
         image.convertTo(QImage::Format_ARGB32);
+        subresDesc.setImage(image);
 
     } else {
-        image = videoFramePlaneAsImage(
-                frame, plane, sizeCompatibleImageFormat(texDesc.textureFormat[plane]), planeSize);
+        // Note, QByteArray::fromRawData creare QByteArray as a view without data copying
+        subresDesc.setData(QByteArray::fromRawData(
+                reinterpret_cast<const char *>(frame.bits(plane)), frame.mappedBytes(plane)));
+        subresDesc.setDataStride(frame.bytesPerLine(plane));
+        result = UpdateTextureWithMapResult::UpdatedWithDataReference;
     }
 
-    if (image.isNull()) {
-        qWarning() << "Cannot represent the plane" << plane << "as an image";
-        return false;
-    }
-
-    subresDesc.setImage(image);
-    subresDesc.setDataStride(image.bytesPerLine());
-
     QRhiTextureUploadEntry entry(0, 0, subresDesc);
     QRhiTextureUploadDescription desc({ entry });
-    rub->uploadTexture(tex.get(), desc);
+    rub.uploadTexture(tex.get(), desc);
 
-    return true;
+    return result;
 }
 
-static std::unique_ptr<QRhiTexture> createTextureFromHandle(const QVideoFrame &frame, QRhi *rhi, int plane)
+static std::unique_ptr<QRhiTexture>
+createTextureFromHandle(QVideoFrameTexturesHandles &texturesSet, QRhi &rhi,
+                        QVideoFrameFormat::PixelFormat pixelFormat, QSize size, int plane)
 {
-    QVideoFrameFormat fmt = frame.surfaceFormat();
-    QVideoFrameFormat::PixelFormat pixelFormat = fmt.pixelFormat();
-    QSize size = fmt.frameSize();
-
     const TextureDescription &texDesc = descriptions[pixelFormat];
-    QSize planeSize(size.width()/texDesc.sizeScale[plane].x, size.height()/texDesc.sizeScale[plane].y);
+    QSize planeSize = texDesc.rhiPlaneSize(size, plane, &rhi);
 
     QRhiTexture::Flags textureFlags = {};
     if (pixelFormat == QVideoFrameFormat::Format_SamplerExternalOES) {
 #ifdef Q_OS_ANDROID
-        if (rhi->backend() == QRhi::OpenGLES2)
+        if (rhi.backend() == QRhi::OpenGLES2)
             textureFlags |= QRhiTexture::ExternalOES;
 #endif
     }
     if (pixelFormat == QVideoFrameFormat::Format_SamplerRect) {
 #ifdef Q_OS_MACOS
-        if (rhi->backend() == QRhi::OpenGLES2)
+        if (rhi.backend() == QRhi::OpenGLES2)
             textureFlags |= QRhiTexture::TextureRectangleGL;
 #endif
     }
 
-    if (quint64 handle = frame.videoBuffer()->textureHandle(plane); handle) {
-        std::unique_ptr<QRhiTexture> tex(rhi->newTexture(texDesc.textureFormat[plane], planeSize, 1, textureFlags));
+    if (quint64 handle = texturesSet.textureHandle(rhi, plane); handle) {
+        std::unique_ptr<QRhiTexture> tex(rhi.newTexture(texDesc.rhiTextureFormat(plane, &rhi), planeSize, 1, textureFlags));
         if (tex->createFrom({handle, 0}))
             return tex;
 
@@ -673,71 +764,106 @@ static std::unique_ptr<QRhiTexture> createTextureFromHandle(const QVideoFrame &f
     return {};
 }
 
-class QVideoFrameTexturesArray : public QVideoFrameTextures
+template <typename TexturesType, typename... Args>
+static QVideoFrameTexturesUPtr
+createTexturesArray(QRhi &rhi, QVideoFrameTexturesHandles &texturesSet,
+                    QVideoFrameFormat::PixelFormat pixelFormat, QSize size, Args &&...args)
 {
-public:
-    using TextureArray = std::array<std::unique_ptr<QRhiTexture>, TextureDescription::maxPlanes>;
-    QVideoFrameTexturesArray(TextureArray &&textures)
-        : m_textures(std::move(textures))
-    {}
-
-    QRhiTexture *texture(uint plane) const override
-    {
-        return plane < std::size(m_textures) ? m_textures[plane].get() : nullptr;
-    }
-
-    TextureArray takeTextures() { return std::move(m_textures); }
-
-private:
-    TextureArray m_textures;
-};
-
-static std::unique_ptr<QVideoFrameTextures> createTexturesFromHandles(const QVideoFrame &frame, QRhi *rhi)
-{
-    const TextureDescription &texDesc = descriptions[frame.surfaceFormat().pixelFormat()];
+    const TextureDescription &texDesc = descriptions[pixelFormat];
     bool ok = true;
-    QVideoFrameTexturesArray::TextureArray textures;
+    RhiTextureArray textures;
     for (quint8 plane = 0; plane < texDesc.nplanes; ++plane) {
-        textures[plane] = QVideoTextureHelper::createTextureFromHandle(frame, rhi, plane);
+        textures[plane] = QVideoTextureHelper::createTextureFromHandle(texturesSet, rhi,
+                                                                       pixelFormat, size, plane);
         ok &= bool(textures[plane]);
     }
     if (ok)
-        return std::make_unique<QVideoFrameTexturesArray>(std::move(textures));
+        return std::make_unique<TexturesType>(std::move(textures), std::forward<Args>(args)...);
     else
         return {};
 }
 
-static std::unique_ptr<QVideoFrameTextures> createTexturesFromMemory(QVideoFrame frame, QRhi *rhi, QRhiResourceUpdateBatch *rub, QVideoFrameTextures *old)
+QVideoFrameTexturesUPtr createTexturesFromHandles(QVideoFrameTexturesHandlesUPtr texturesSet,
+                                                  QRhi &rhi,
+                                                  QVideoFrameFormat::PixelFormat pixelFormat,
+                                                  QSize size)
 {
-    const TextureDescription &texDesc = descriptions[frame.surfaceFormat().pixelFormat()];
-    QVideoFrameTexturesArray::TextureArray textures;
-    auto oldArray = dynamic_cast<QVideoFrameTexturesArray *>(old);
-    if (oldArray)
-        textures = oldArray->takeTextures();
+    if (!texturesSet)
+        return nullptr;
 
-    bool ok = true;
-    for (quint8 plane = 0; plane < texDesc.nplanes; ++plane) {
-        ok &= updateTextureWithMap(frame, rhi, rub, plane, textures[plane]);
+    if (pixelFormat == QVideoFrameFormat::Format_Invalid)
+        return nullptr;
+
+    if (size.isEmpty())
+        return nullptr;
+
+    auto &texturesSetRef = *texturesSet;
+    return createTexturesArray<QVideoFrameTexturesFromHandlesSet>(rhi, texturesSetRef, pixelFormat,
+                                                                  size, std::move(texturesSet));
+}
+
+static QVideoFrameTexturesUPtr createTexturesFromMemory(QVideoFrame frame, QRhi &rhi,
+                                                        QRhiResourceUpdateBatch &rub,
+                                                        QVideoFrameTexturesUPtr &oldTextures)
+{
+    if (!frame.map(QVideoFrame::ReadOnly)) {
+        qWarning() << "Cannot map a video frame in ReadOnly mode!";
+        return {};
     }
-    if (ok)
-        return std::make_unique<QVideoFrameTexturesArray>(std::move(textures));
-    else
-        return {};
+
+    auto unmapFrameGuard = qScopeGuard([&frame] { frame.unmap(); });
+
+    const TextureDescription &texDesc = descriptions[frame.surfaceFormat().pixelFormat()];
+
+    const bool canReuseTextures(dynamic_cast<QVideoFrameTexturesFromMemory*>(oldTextures.get()));
+
+    std::unique_ptr<QVideoFrameTexturesFromMemory> textures(canReuseTextures ?
+                static_cast<QVideoFrameTexturesFromMemory *>(oldTextures.release()) :
+                new QVideoFrameTexturesFromMemory);
+
+    RhiTextureArray& textureArray = textures->textureArray();
+    bool shouldKeepMapping = false;
+    for (quint8 plane = 0; plane < texDesc.nplanes; ++plane) {
+        const auto result = updateTextureWithMap(frame, rhi, rub, plane, textureArray[plane]);
+        if (result == UpdateTextureWithMapResult::Failed)
+            return {};
+
+        if (result == UpdateTextureWithMapResult::UpdatedWithDataReference)
+            shouldKeepMapping = true;
+    }
+
+    // as QVideoFrame::unmap does nothing with null frames, we just move the frame to the result
+    textures->setMappedFrame(shouldKeepMapping ? std::move(frame) : QVideoFrame());
+
+    return textures;
 }
 
-std::unique_ptr<QVideoFrameTextures> createTextures(QVideoFrame &frame, QRhi *rhi, QRhiResourceUpdateBatch *rub, std::unique_ptr<QVideoFrameTextures> &&oldTextures)
+QVideoFrameTexturesUPtr createTextures(const QVideoFrame &frame, QRhi &rhi,
+                                       QRhiResourceUpdateBatch &rub,
+                                       QVideoFrameTexturesUPtr oldTextures)
 {
-    QAbstractVideoBuffer *vf = frame.videoBuffer();
-    if (!vf)
+    if (!frame.isValid())
         return {};
 
-    if (auto vft = vf->mapTextures(rhi))
-        return vft;
+    auto setSourceFrame = [&frame](QVideoFrameTexturesUPtr result) {
+        result->setSourceFrame(frame);
+        return result;
+    };
 
-    if (auto vft = createTexturesFromHandles(frame, rhi))
-        return vft;
+    if (QHwVideoBuffer *hwBuffer = QVideoFramePrivate::hwBuffer(frame)) {
+        if (auto textures = hwBuffer->mapTextures(rhi, oldTextures))
+            return setSourceFrame(std::move(textures));
 
-    return createTexturesFromMemory(frame, rhi, rub, oldTextures.get());
+        QVideoFrameFormat format = frame.surfaceFormat();
+        if (auto textures = createTexturesArray<QVideoFrameTexturesFromRhiTextureArray>(
+                    rhi, *hwBuffer, format.pixelFormat(), format.frameSize()))
+            return setSourceFrame(std::move(textures));
+    }
+
+    if (auto textures = createTexturesFromMemory(frame, rhi, rub, oldTextures))
+        return setSourceFrame(std::move(textures));
+
+    return {};
 }
 
 bool SubtitleLayout::update(const QSize &frameSize, QString text)
