@@ -284,11 +284,12 @@ static QMutex sessionMutex;
     m_player = [AVPlayer playerWithPlayerItem:m_playerItem];
     [m_player retain];
 
-    //Set the initial volume on new player object
+    //Set the initial audio ouptut settings on new player object
     if (self.session) {
         auto *audioOutput = m_session->m_audioOutput;
         m_player.volume = (audioOutput ? audioOutput->volume : 1.);
         m_player.muted = (audioOutput ? audioOutput->muted : true);
+        m_session->updateAudioOutputDevice();
     }
 
     //Assign the output layer to the new player
@@ -441,6 +442,9 @@ static QMutex sessionMutex;
 
     [m_mimeType release];
     [m_playerLayer release];
+    // 'videoTrack' is a 'retain' property, but still needs a
+    // manual 'release' (i.e. setting to nil):
+    self.videoTrack = nil;
     [super dealloc];
 }
 
@@ -465,10 +469,12 @@ static QMutex sessionMutex;
     if (loadingRequest.dataRequest) {
         NSInteger requestedLength = loadingRequest.dataRequest.requestedLength;
         int maxBytes = qMin(32 * 1064, int(requestedLength));
-        char buffer[maxBytes];
+        QByteArray buffer;
+        buffer.resize(maxBytes);
+
         NSInteger submitted = 0;
         while (submitted < requestedLength) {
-            qint64 len = device->read(buffer, maxBytes);
+            qint64 len = device->read(buffer.data(), maxBytes);
             if (len < 1)
                 break;
 
@@ -596,10 +602,10 @@ void AVFMediaPlayer::setMedia(const QUrl &content, QIODevice *stream)
     setSeekable(false);
     m_requestedPosition = -1;
     orientationChanged(QtVideo::Rotation::None, false);
-    Q_EMIT positionChanged(position());
+    positionChanged(position());
     if (m_duration != 0) {
         m_duration = 0;
-        Q_EMIT durationChanged(0);
+        durationChanged(0);
     }
     if (!m_metaData.isEmpty()) {
         m_metaData.clear();
@@ -617,18 +623,18 @@ void AVFMediaPlayer::setMedia(const QUrl &content, QIODevice *stream)
     if (!m_mediaStream && content.isEmpty()) {
         m_mediaStatus = QMediaPlayer::NoMedia;
         if (m_mediaStatus != oldMediaStatus)
-            Q_EMIT mediaStatusChanged(m_mediaStatus);
+            mediaStatusChanged(m_mediaStatus);
 
         m_state = QMediaPlayer::StoppedState;
         if (m_state != oldState)
-            Q_EMIT stateChanged(m_state);
+            stateChanged(m_state);
 
         return;
     }
 
     m_mediaStatus = QMediaPlayer::LoadingMedia;
     if (m_mediaStatus != oldMediaStatus)
-        Q_EMIT mediaStatusChanged(m_mediaStatus);
+        mediaStatusChanged(m_mediaStatus);
 
     if (m_mediaStream) {
         // If there is a data, try to load it,
@@ -643,7 +649,7 @@ void AVFMediaPlayer::setMedia(const QUrl &content, QIODevice *stream)
 
     m_state = QMediaPlayer::StoppedState;
     if (m_state != oldState)
-        Q_EMIT stateChanged(m_state);
+        stateChanged(m_state);
 }
 
 qint64 AVFMediaPlayer::position() const
@@ -682,7 +688,7 @@ void AVFMediaPlayer::setAudioAvailable(bool available)
         return;
 
     m_audioAvailable = available;
-    Q_EMIT audioAvailableChanged(available);
+    audioAvailableChanged(available);
 }
 
 bool AVFMediaPlayer::isAudioAvailable() const
@@ -696,7 +702,7 @@ void AVFMediaPlayer::setVideoAvailable(bool available)
         return;
 
     m_videoAvailable = available;
-    Q_EMIT videoAvailableChanged(available);
+    videoAvailableChanged(available);
 }
 
 bool AVFMediaPlayer::isVideoAvailable() const
@@ -750,12 +756,12 @@ void AVFMediaPlayer::setAudioOutput(QPlatformAudioOutput *output)
         m_audioOutput->q->disconnect(this);
     m_audioOutput = output;
     if (m_audioOutput) {
-        connect(m_audioOutput->q, &QAudioOutput::deviceChanged, this, &AVFMediaPlayer::audioOutputChanged);
+        connect(m_audioOutput->q, &QAudioOutput::deviceChanged, this, &AVFMediaPlayer::updateAudioOutputDevice);
         connect(m_audioOutput->q, &QAudioOutput::volumeChanged, this, &AVFMediaPlayer::setVolume);
         connect(m_audioOutput->q, &QAudioOutput::mutedChanged, this, &AVFMediaPlayer::setMuted);
         //connect(m_audioOutput->q, &QAudioOutput::audioRoleChanged, this, &AVFMediaPlayer::setAudioRole);
     }
-    audioOutputChanged();
+    updateAudioOutputDevice();
     setMuted(m_audioOutput ? m_audioOutput->muted : true);
     setVolume(m_audioOutput ? m_audioOutput->volume : 1.);
 }
@@ -780,7 +786,7 @@ void AVFMediaPlayer::setPlaybackRate(qreal rate)
     if (player && m_state == QMediaPlayer::PlayingState)
         [player setRate:m_rate];
 
-    Q_EMIT playbackRateChanged(m_rate);
+    playbackRateChanged(m_rate);
 }
 
 void AVFMediaPlayer::setPosition(qint64 pos)
@@ -795,14 +801,14 @@ void AVFMediaPlayer::setPosition(qint64 pos)
     AVPlayerItem *playerItem = [static_cast<AVFMediaPlayerObserver*>(m_observer) playerItem];
     if (!playerItem) {
         m_requestedPosition = pos;
-        Q_EMIT positionChanged(m_requestedPosition);
+        positionChanged(m_requestedPosition);
         return;
     }
 
     if (!isSeekable()) {
         if (m_requestedPosition != -1) {
             m_requestedPosition = -1;
-            Q_EMIT positionChanged(position());
+            positionChanged(position());
         }
         return;
     }
@@ -820,13 +826,13 @@ void AVFMediaPlayer::setPosition(qint64 pos)
                                     m_requestedPosition = -1;
                            }];
 
-    Q_EMIT positionChanged(pos);
+    positionChanged(pos);
 
     // Reset media status if the current status is EndOfMedia
     if (m_mediaStatus == QMediaPlayer::EndOfMedia) {
         QMediaPlayer::MediaStatus newMediaStatus = (m_state == QMediaPlayer::PausedState) ? QMediaPlayer::BufferedMedia
                                                                                           : QMediaPlayer::LoadedMedia;
-        Q_EMIT mediaStatusChanged((m_mediaStatus = newMediaStatus));
+        mediaStatusChanged((m_mediaStatus = newMediaStatus));
     }
 }
 
@@ -859,7 +865,7 @@ void AVFMediaPlayer::play()
     m_state = QMediaPlayer::PlayingState;
     processLoadStateChange();
 
-    Q_EMIT stateChanged(m_state);
+    stateChanged(m_state);
     m_playbackTimer.start(100);
 }
 
@@ -886,8 +892,8 @@ void AVFMediaPlayer::pause()
     if (m_mediaStatus == QMediaPlayer::EndOfMedia)
         setPosition(0);
 
-    Q_EMIT positionChanged(position());
-    Q_EMIT stateChanged(m_state);
+    positionChanged(position());
+    stateChanged(m_state);
     m_playbackTimer.stop();
 }
 
@@ -908,9 +914,9 @@ void AVFMediaPlayer::stop()
         m_videoOutput->setLayer(nullptr);
 
     if (m_mediaStatus == QMediaPlayer::BufferedMedia)
-        Q_EMIT mediaStatusChanged((m_mediaStatus = QMediaPlayer::LoadedMedia));
+        mediaStatusChanged((m_mediaStatus = QMediaPlayer::LoadedMedia));
 
-    Q_EMIT stateChanged((m_state = QMediaPlayer::StoppedState));
+    stateChanged((m_state = QMediaPlayer::StoppedState));
     m_playbackTimer.stop();
 }
 
@@ -936,14 +942,17 @@ void AVFMediaPlayer::setMuted(bool muted)
         player.muted = muted;
 }
 
-void AVFMediaPlayer::audioOutputChanged()
+void AVFMediaPlayer::updateAudioOutputDevice()
 {
 #ifdef Q_OS_MACOS
     AVPlayer *player = [static_cast<AVFMediaPlayerObserver*>(m_observer) player];
+    if (!player)
+        return;
+
     if (!m_audioOutput || m_audioOutput->device.id().isEmpty()) {
-        player.audioOutputDeviceUniqueID = nil;
         if (!m_audioOutput)
             player.muted = true;
+        player.audioOutputDeviceUniqueID = nil;
     } else {
         NSString *str = QString::fromUtf8(m_audioOutput->device.id()).toNSString();
         player.audioOutputDeviceUniqueID = str;
@@ -963,15 +972,15 @@ void AVFMediaPlayer::processEOS()
 #ifdef QT_DEBUG_AVF
     qDebug() << Q_FUNC_INFO;
 #endif
-    Q_EMIT positionChanged(position());
+    positionChanged(position());
     m_mediaStatus = QMediaPlayer::EndOfMedia;
     m_state = QMediaPlayer::StoppedState;
 
     if (m_videoOutput)
         m_videoOutput->setLayer(nullptr);
 
-    Q_EMIT mediaStatusChanged(m_mediaStatus);
-    Q_EMIT stateChanged(m_state);
+    mediaStatusChanged(m_mediaStatus);
+    stateChanged(m_state);
 }
 
 void AVFMediaPlayer::processLoadStateChange(QMediaPlayer::PlaybackState newState)
@@ -993,7 +1002,7 @@ void AVFMediaPlayer::processLoadStateChange(QMediaPlayer::PlaybackState newState
 
         // get the meta data
         m_metaData = AVFMetaData::fromAsset(playerItem.asset);
-        Q_EMIT metaDataChanged();
+        metaDataChanged();
         updateTracks();
 
         if (playerItem) {
@@ -1019,8 +1028,7 @@ void AVFMediaPlayer::processLoadStateChange(QMediaPlayer::PlaybackState newState
                                                              : QMediaPlayer::LoadedMedia;
 
         if (newStatus != m_mediaStatus)
-            Q_EMIT mediaStatusChanged((m_mediaStatus = newStatus));
-
+            mediaStatusChanged((m_mediaStatus = newStatus));
     }
 
     if (newState == QMediaPlayer::PlayingState && [static_cast<AVFMediaPlayerObserver*>(m_observer) player]) {
@@ -1039,7 +1047,7 @@ void AVFMediaPlayer::processLoadStateChange()
 
 void AVFMediaPlayer::processLoadStateFailure()
 {
-    Q_EMIT stateChanged((m_state = QMediaPlayer::StoppedState));
+    stateChanged((m_state = QMediaPlayer::StoppedState));
 }
 
 void AVFMediaPlayer::processBufferStateChange(int bufferProgress)
@@ -1061,10 +1069,10 @@ void AVFMediaPlayer::processBufferStateChange(int bufferProgress)
     }
 
     if (m_mediaStatus != status)
-        Q_EMIT mediaStatusChanged(m_mediaStatus = status);
+        mediaStatusChanged(m_mediaStatus = status);
 
     m_bufferProgress = bufferProgress;
-    Q_EMIT bufferProgressChanged(bufferProgress/100.);
+    bufferProgressChanged(bufferProgress / 100.);
 }
 
 void AVFMediaPlayer::processDurationChange(qint64 duration)
@@ -1073,7 +1081,7 @@ void AVFMediaPlayer::processDurationChange(qint64 duration)
         return;
 
     m_duration = duration;
-    Q_EMIT durationChanged(duration);
+    durationChanged(duration);
 }
 
 void AVFMediaPlayer::processPositionChange()
@@ -1081,19 +1089,19 @@ void AVFMediaPlayer::processPositionChange()
     if (m_state == QMediaPlayer::StoppedState)
         return;
 
-    Q_EMIT positionChanged(position());
+    positionChanged(position());
 }
 
 void AVFMediaPlayer::processMediaLoadError()
 {
     if (m_requestedPosition != -1) {
         m_requestedPosition = -1;
-        Q_EMIT positionChanged(position());
+        positionChanged(position());
     }
 
-    Q_EMIT mediaStatusChanged((m_mediaStatus = QMediaPlayer::InvalidMedia));
+    mediaStatusChanged((m_mediaStatus = QMediaPlayer::InvalidMedia));
 
-    Q_EMIT error(QMediaPlayer::FormatError, tr("Failed to load media"));
+    error(QMediaPlayer::FormatError, tr("Failed to load media"));
 }
 
 void AVFMediaPlayer::streamReady()
@@ -1151,7 +1159,7 @@ void AVFMediaPlayer::updateTracks()
         if (firstLoad)
             setActiveTrack(SubtitleStream, -1);
     }
-    Q_EMIT tracksChanged();
+    tracksChanged();
 }
 
 void AVFMediaPlayer::setActiveTrack(QPlatformMediaPlayer::TrackType type, int index)
@@ -1165,17 +1173,29 @@ void AVFMediaPlayer::setActiveTrack(QPlatformMediaPlayer::TrackType type, int in
             AVAsset *asset = playerItem.asset;
             if (!asset)
                 return;
+#if defined(Q_OS_VISIONOS)
+            [asset loadMediaSelectionGroupForMediaCharacteristic:AVMediaCharacteristicLegible
+                                               completionHandler:[=](AVMediaSelectionGroup *group, NSError *error) {
+                                                   // FIXME: handle error
+                                                   if (error)
+                                                       return;
+                                                   auto *options = group.options;
+                                                   if (options.count)
+                                                       [playerItem selectMediaOption:options.firstObject inMediaSelectionGroup:group];
+                                               }];
+#else
             AVMediaSelectionGroup *group = [asset mediaSelectionGroupForMediaCharacteristic:AVMediaCharacteristicLegible];
             if (!group)
                 return;
             auto *options = group.options;
             if (options.count)
                 [playerItem selectMediaOption:options.firstObject inMediaSelectionGroup:group];
+#endif
         }
     }
     for (int i = 0; i < t.count(); ++i)
         t.at(i).enabled = (i == index);
-    emit activeTracksChanged();
+    activeTracksChanged();
 }
 
 int AVFMediaPlayer::activeTrack(QPlatformMediaPlayer::TrackType type)
